/**
 *  Copyright 2011 Carsten Gräf
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 * 
 */

package de.twenty11.skysail.common.filters;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import de.twenty11.skysail.common.internal.filters.FilterExpressionParser;

/**
 * A LdapSearchFilter is the combination of a type, an attribute and a value.
 * 
 * The value can be a list of further filters.
 * 
 * @author carsten
 * 
 */
public class LdapSearchFilter extends AbstractFilter {

    /**
     * 
     */
    private static final long serialVersionUID = 6320587880866948360L;

    /**
     * Create a LdapSearchFilter object from a filter expression
     * 
     * @param filterString
     *            a filter expression like "(topic=OSGi)"
     * @return
     */
    public static Filter newInstance(String filterString) {
        return new FilterExpressionParser(filterString).parse();
    }

    // /**
    // * Create a LdapSearchFilter object from a Map of String to String. Please
    // note that
    // * all filter entries will automatically be "AND"ed.
    // *
    // * @param valuesMap
    // */
    // public static Filter newInstance(Map<String, String> valuesMap) {
    // if (valuesMap == null || valuesMap.size() == 0) {
    // return new LdapSearchFilter(FilterType.EMPTY);
    // }
    // List<LdapSearchFilter> operands = new ArrayList<LdapSearchFilter>(10);
    // for (String key : valuesMap.keySet()) {
    // // Create one filter for each key with non-empty value
    // String value = valuesMap.get(key);
    // if (value != null) {
    // LdapSearchFilter filter = new LdapSearchFilter(FilterType.EQUALITY_MATCH,
    // key, valuesMap.get(key));
    // operands.add(filter);
    // }
    // }
    // return new LdapSearchFilter(FilterType.AND, null, operands);
    // }

    /**
     * 
     * @param type
     */
    // TODO ugly
    public LdapSearchFilter(FilterType type) {
        setFilterType(type);
        setAttribute(null);
        setSubfilters(null);
        setValue(null);
    }

    public LdapSearchFilter(FilterType present, String attribute) {
        setAttribute(attribute);
        setSubfilters(null);
    }

    public LdapSearchFilter(FilterType type, String attribute, Filter subfilter) {
        setFilterType(type);
        setAttribute(attribute);

        List<Filter> singleEntryList = new ArrayList<Filter>();
        singleEntryList.add(subfilter);
        setSubfilters(singleEntryList);
        setValue(null);
    }

    // TODO make private? new Instance?
    public LdapSearchFilter(FilterType type, String attribute, List<Filter> subfilters) {
        setFilterType(type);
        setAttribute(attribute);
        setSubfilters(subfilters);
        setValue(null);
    }

    public LdapSearchFilter(FilterType type, String attr, Object object) {
        setFilterType(type);
        setAttribute(attr);
        setSubfilters(null);
        setValue(object.toString());
    }

    public LdapSearchFilter(FilterType type, String attribute, String value) {
        setFilterType(type);
        setAttribute(attribute);
        setSubfilters(null);
        setValue(value);
    }

    @Override
    public String getValue(String attribute) {
        return super.getValue();
    }

    @Override
    public boolean match(Map<String, String> properties) {
        switch (getFilterType()) {
        case EMPTY: {
            return true;
        }
        case AND: {
            for (Filter filter : getSubfilters()) {
                if (!filter.match(properties)) {
                    return false;
                }
            }
            return true;
        }

        case OR: {
            for (Filter filter : getSubfilters()) {
                if (filter.match(properties)) {
                    return true;
                }
            }
            return false;
        }

        case NOT: {
            Filter filter = getSubfilters().get(0);
            return !filter.match(properties);
        }
        case PRESENT: {
            Object prop = properties == null ? null : properties.get(getAttribute());
            return prop != null;
        }
        default:
            return false;
        }
    }

    private String print() {
        StringBuffer sb = new StringBuffer();
        sb.append('(');

        switch (getFilterType()) {
        case AND: {
            sb.append('&');
            for (Filter filter : getSubfilters()) {
                sb.append(filter.toString());
            }
            break;
        }
        case OR: {
            sb.append('|');
            for (Filter filter : getSubfilters()) {
                sb.append(filter.toString());
            }
            break;
        }
        case NOT: {
            sb.append('!');
            sb.append(getSubfilters().get(0).toString());
            break;
        }
        case EQUALITY_MATCH: {
            sb.append(getAttribute());
            sb.append('=');
            sb.append(getValue());

            break;
        }
        case GREATER_OR_EQUAL: {
            sb.append(getAttribute());
            sb.append(">="); //$NON-NLS-1$
            sb.append(getValue());

            break;
        }
        case LESS_OR_EQUAL: {
            sb.append(getAttribute());
            sb.append("<="); //$NON-NLS-1$
            sb.append(getValue());
            break;
        }
        case PRESENT: {
            sb.append(getAttribute());
            sb.append("=*"); //$NON-NLS-1$
            break;
        }
        default:
        }

        sb.append(')');

        return sb.toString();
    }

    @Override
    public String toString() {
        return print();
    }

}
