/**
 *  Copyright 2011 Carsten Gräf
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 * 
 */

package de.twenty11.skysail.common.internal.filters;

import java.util.ArrayList;
import java.util.List;

import de.twenty11.skysail.common.filters.Filter;
import de.twenty11.skysail.common.filters.FilterType;
import de.twenty11.skysail.common.filters.LdapSearchFilter;

public class FilterExpressionParser {

    private static final String COULD_NOT_PARSE_FILTER_EXPRESSION = "could not parse filter expression ";

    /** the string representation of the filter */
    private final String filterExpression;

    /** list of characters from the filter Expression */
    private final char[] filterChars;

    /**
     * position of "cursor" in the filter characters, starting with 1 (first char expected to be '(')
     */
    private int position = 1;

    public FilterExpressionParser(String filter) {
        if (filter == null) {
            throw new IllegalArgumentException("filter expression is null");
        }
        if (filter.trim().length() == 0) {
            throw new IllegalArgumentException("filter expression is empty");
        }
        // remove all whitespace characters first
        filterExpression = filter.replaceAll("\\s+", "");
        filterChars = filterExpression.toCharArray();
    }

    public LdapSearchFilter parse() {
        // constraints
        if (!(filterExpression.charAt(0) == '(')) {
            throw new RuntimeException("filter expression '" + filterExpression + "' doesn't start with '('");
        }
        if (!filterExpression.endsWith(")")) {
            throw new RuntimeException("filter expression '" + filterExpression + "' doesn't end with ')'");
        }
        // naming see http://www.ietf.org/rfc/rfc1960.txt 3.)
        return parseFiltercomp();
    }

    private LdapSearchFilter parseFiltercomp() {
        switch (filterChars[position]) {
        case '&': {
            position++;
            return parseAnd();
        }
        case '|': {
            position++;
            return parseOr();
        }
        case '!': {
            position++;
            return parseNot();
        }
        default:
        }
        return parseItem();
    }

    private LdapSearchFilter parseAnd() {
        List<Filter> subfilters = new ArrayList<Filter>(10);
        while (filterChars[position] == '(') {
            Filter child = parse();
            subfilters.add(child);
        }
        return new LdapSearchFilter(FilterType.AND, null, subfilters);
    }

    private LdapSearchFilter parseOr() {
        List<Filter> subfilters = new ArrayList<Filter>(10);
        while (filterChars[position] == '(') {
            Filter child = parse();
            subfilters.add(child);
        }
        return new LdapSearchFilter(FilterType.OR, null, subfilters);
    }

    private LdapSearchFilter parseNot() {
        LdapSearchFilter child = parse();
        return new LdapSearchFilter(FilterType.NOT, null, child);
    }

    private LdapSearchFilter parseItem() {
        String attr = parseAttribute();

        switch (filterChars[position]) {
        case '>': {
            if (filterChars[position + 1] == '=') {
                position += 2;
                return new LdapSearchFilter(FilterType.GREATER_OR_EQUAL, attr, parseValue());
            }
            break;
        }
        case '<': {
            if (filterChars[position + 1] == '=') {
                position += 2;
                return new LdapSearchFilter(FilterType.LESS_OR_EQUAL, attr, parseValue());
            }
            break;
        }
        case '=': {
            if (filterChars[position + 1] == '*') {
                int oldpos = position;
                position += 2;

                if (filterChars[position] == ')') {
                    return new LdapSearchFilter(FilterType.PRESENT, attr);
                }
                position = oldpos;
            }

            position++;
            Object string = parseSubstring();

            if (string instanceof String) {
                return new LdapSearchFilter(FilterType.EQUALITY_MATCH, attr, (String) string);
            }
            return new LdapSearchFilter(FilterType.SUBSTRINGS, attr, string);
        }
        default:
            throw new RuntimeException("unknown attribute '" + filterChars[position] + "' for LdapSearchFilter");
        }

        throw new RuntimeException(COULD_NOT_PARSE_FILTER_EXPRESSION + filterExpression);
    }

    private String parseAttribute() {

        int begin = position;
        int end = position;

        char c = filterChars[position];

        while (c != '~' && c != '<' && c != '>' && c != '=' && c != '(' && c != ')') {
            position++;

            if (!Character.isWhitespace(c)) {
                end = position;
            }

            c = filterChars[position];
        }

        int length = end - begin;

        if (length == 0) {
            throw new RuntimeException(COULD_NOT_PARSE_FILTER_EXPRESSION + filterExpression);
        }

        return new String(filterChars, begin, length);
    }

    private String parseValue() {
        StringBuffer sb = new StringBuffer(filterChars.length - position);

        parseloop: while (true) {
            char c = filterChars[position];

            switch (c) {
            case ')': {
                break parseloop;
            }

            case '(': {
                throw new RuntimeException(COULD_NOT_PARSE_FILTER_EXPRESSION + filterExpression);
            }

            case '\\': {
                position++;
                c = filterChars[position];
                sb.append(c);
                position++;
                break;
            }

            default: {
                sb.append(c);
                position++;
                break;
            }
            }
        }

        if (sb.length() == 0) {
            throw new RuntimeException(COULD_NOT_PARSE_FILTER_EXPRESSION + filterExpression);
        }

        return sb.toString();
    }

    private Object parseSubstring() {
        StringBuffer sb = new StringBuffer(filterChars.length - position);

        List operands = new ArrayList(10);

        parseloop: while (true) {
            char c = filterChars[position];

            switch (c) {
            case ')': {
                if (sb.length() > 0) {
                    operands.add(sb.toString());
                }

                break parseloop;
            }

            case '(': {
                throw new RuntimeException(COULD_NOT_PARSE_FILTER_EXPRESSION + filterExpression);
            }

            case '*': {
                if (sb.length() > 0) {
                    operands.add(sb.toString());
                }

                sb.setLength(0);

                operands.add(null);
                position++;

                break;
            }

            case '\\': {
                position++;
                c = filterChars[position];
                sb.append(c);
                position++;
                break;
            }

            default: {
                sb.append(c);
                position++;
                break;
            }
            }
        }

        int size = operands.size();

        if (size == 0) {
            return "";
        }

        if (size == 1) {
            Object single = operands.get(0);

            if (single != null) {
                return single;
            }
        }

        return operands.toArray(new String[size]);
    }

}
